import os
import shutil
import tempfile
import unittest
from unittest import mock
from zipfile import ZipFile

from shub import deploy_egg
from shub.exceptions import BadParameterException


class FakeRequester:
    """Used to mock shub.utils#make_deploy_request"""
    def fake_request(self, *args):
        self.url = args[0]
        self.data = args[1]
        self.files = args[2]
        self.auth = args[3]


@mock.patch.dict(os.environ, {'SHUB_APIKEY': '1234'})
class TestDeployEgg(unittest.TestCase):

    def setUp(self):
        self.curdir = os.getcwd()
        self.fake_requester = FakeRequester()
        deploy_egg.utils.make_deploy_request = self.fake_requester.fake_request
        self.tmp_dir = tempfile.mkdtemp(prefix="shub-test-deploy-eggs")

    def tearDown(self):
        os.chdir(self.curdir)
        if os.path.exists(self.tmp_dir):
            shutil.rmtree(self.tmp_dir)

    def test_parses_project_information_correctly(self):
        # this test's assertions are based on the values
        # defined on this folder's setup.py file
        shutil.rmtree(self.tmp_dir)
        shutil.copytree('tests/samples/deploy_egg_sample_project', self.tmp_dir)
        os.chdir(self.tmp_dir)

        data = self.call_main_and_check_request_data()
        self.assertEqual('1.2.0', data['version'])

    def test_can_clone_a_git_repo_and_deploy_the_egg(self):
        self._unzip_git_repo_to(self.tmp_dir)
        repo = os.path.join(self.tmp_dir, 'deploy_egg_sample_repo.git')

        self.call_main_and_check_request_data(from_url=repo)
        data = self.call_main_and_check_request_data()

        self.assertTrue('master' in data['version'])

    @unittest.skip('flaky')
    def test_can_deploy_an_egg_from_pypi(self):
        basepath = os.path.abspath('tests/samples/')
        pkg = os.path.join(basepath, 'deploy_egg_sample_project.zip')
        self.call_main_and_check_request_data(from_pypi=pkg)

    def test_can_clone_checkout_and_deploy_the_egg(self):
        self._unzip_git_repo_to(self.tmp_dir)
        repo = os.path.join(self.tmp_dir, 'deploy_egg_sample_repo.git')

        branch = 'dev'
        data = self.call_main_and_check_request_data(from_url=repo, git_branch=branch)
        self.assertTrue('dev' in data['version'])

    def test_fails_on_invalid_repo(self):
        self._unzip_git_repo_to(self.tmp_dir)
        repo = os.path.join(self.tmp_dir, 'deploy_egg_sample_repo.git')
        shutil.rmtree(os.path.join(repo, '.git'))

        with self.assertRaises(BadParameterException):
            self.call_main_and_check_request_data(from_url=repo)

    def test_fails_on_invalid_branch(self):
        self._unzip_git_repo_to(self.tmp_dir)
        repo = os.path.join(self.tmp_dir, 'deploy_egg_sample_repo.git')
        with self.assertRaises(BadParameterException):
            self.call_main_and_check_request_data(
                from_url=repo, git_branch='nonexisting')

    def _unzip_git_repo_to(self, path):
        zipped_repo = os.path.abspath('tests/samples/deploy_egg_sample_repo.git.zip')
        ZipFile(zipped_repo).extractall(path)

    def call_main_and_check_request_data(self, project_id=0, from_url=None,
                                         git_branch=None, from_pypi=None):
        # WHEN
        deploy_egg.main(project_id, from_url, git_branch, from_pypi)

        data = self.fake_requester.data
        files = self.fake_requester.files

        # THEN
        # the egg was successfully built, let's check the data
        # that is sent to the scrapy cloud
        self.assertTrue('test_project', files['egg'][0])
        self.assertEqual(project_id, data['project'])
        self.assertEqual('test_project', data['name'])

        return data



import os
import shutil
import tempfile
import zipfile

import click
import requests
import yaml
from click.formatting import HelpFormatter
from io import BytesIO

from shub.exceptions import (
    BadParameterException, NotFoundException, RemoteErrorException)


EXAMPLE_REPO = "scrapinghub/custom-images-examples"
AVAILABLE_PROJECTS_URL = (
    "https://raw.githubusercontent.com/%s/master/bootstrap_projects.yml"
    "" % EXAMPLE_REPO)

HELP = """
Through custom images, Scrapinghub allows you to run crawlers written in any
language you want. To get you started, we prepared a few examples projects in
different programming languages and frameworks. You can find them in our custom
images repository at:

    https://github.com/scrapinghub/custom-images-examples

The 'shub bootstrap' command clones an example project to the current directory
so that you can start hacking right away.

Run

    shub bootstrap -l

to get a list of all available example projects, then run

    shub bootstrap PROJECTNAME

to clone it.
"""

SHORT_HELP = "Clone custom image example project"


def list_projects_callback(ctx, param, value):
    if not value or ctx.resilient_parsing:
        return
    projects = get_available_projects()
    list_projects(projects)
    ctx.exit()


@click.command(help=HELP, short_help=SHORT_HELP)
@click.option('-l', '--list', 'list_projects', help='list available projects',
              is_flag=True, callback=list_projects_callback,
              expose_value=False, is_eager=True)
@click.argument('project')
@click.argument('target_dir', required=False)
def cli(project, target_dir):
    target_dir = os.path.normpath(
        os.path.join(os.getcwd(), target_dir or project))
    if os.path.exists(target_dir):
        raise BadParameterException(
            "Target directory %s already exists, please delete it or supply a "
            "non-existing target." % target_dir)
    projects = get_available_projects()
    if project not in projects:
        raise NotFoundException(
            "There is no example project named '%s'. Run 'shub bootstrap -l' "
            "to get a list of all available projects." % project)
    click.echo("Downloading custom image examples")
    repo_zip = get_repo_zip(EXAMPLE_REPO)
    click.echo(f"Cloning project '{project}' into {target_dir}")
    unzip_project(repo_zip, project=projects[project], target_dir=target_dir)


def get_available_projects():
    try:
        resp = requests.get(AVAILABLE_PROJECTS_URL)
        resp.raise_for_status()
    except (requests.HTTPError, requests.ConnectionError) as e:
        raise RemoteErrorException(
            "There was an error while getting the list of available projects "
            "from GitHub: %s.\n\nPlease check your connection or go to\n  %s\n"
            "to browse the custom image examples manually."
            "" % (e, "https://github.com/%s" % EXAMPLE_REPO))
    return yaml.safe_load(resp.text)


def list_projects(projects):
    formatter = HelpFormatter()
    with formatter.section("Available projects"):
        formatter.write_dl(
            sorted(
                [(name, info['description'])
                 for name, info in projects.items()],
                key=lambda x: x[0]))
    click.echo(formatter.getvalue().strip())


def get_repo_zip(repo):
    zip_url = "https://github.com/%s/archive/master.zip" % repo
    resp = requests.get(zip_url)
    return zipfile.ZipFile(BytesIO(resp.content))


def unzip_project(repo_zip, project, target_dir):
    filenames = repo_zip.namelist()
    repo_dirname = filenames[0]
    project_filenames = [
        fn
        for fn in filenames
        if fn.startswith(repo_dirname + project['path'])
    ]
    tempdir = tempfile.mkdtemp()
    repo_zip.extractall(path=tempdir, members=project_filenames)
    shutil.move(
        os.path.join(tempdir, repo_dirname, project['path']),
        target_dir,
    )
    shutil.rmtree(tempdir)



import os
import re
import ast
import json
import time
import textwrap

import click
import requests
from retrying import retry
from urllib.parse import urljoin

from shub.config import load_shub_config, list_targets_callback
from shub.exceptions import ShubException
from shub.image import utils
from shub.image import list as list_mod


VALIDSPIDERNAME = re.compile('^[a-z0-9][-._a-z0-9]+$', re.I)
STORE_N_LAST_STATUS_URLS = 5
SYNC_DEPLOY_REFRESH_TIMEOUT = 1
SYNC_DEPLOY_WAIT_STATUSES = ['pending', 'started', 'retry', 'progress']
SHORT_HELP = 'Deploy a release image to Scrapy cloud'
CHECK_RETRY_EXCEPTIONS = (
    requests.exceptions.Timeout,
    requests.exceptions.ConnectionError,
    requests.exceptions.HTTPError,
)
# Exponential retry timeouts: min(2^n * multiplier, max)
# [2s, 4s, 8s, 16s, 30s] = 60s
CHECK_RETRY_ATTEMPTS = 6
CHECK_RETRY_EXP_MULTIPLIER = 1000
CHECK_RETRY_EXP_MAX = 30000
DEFAULT_TOTAL_PROGRESS = 100

HELP = """
A command to deploy your release image to Scrapy Cloud.
Does a simple POST request to Dash API with given parameters
(some params are extracted from the project repo, i.e. spiders/scripts).
"""


@click.command(help=HELP, short_help=SHORT_HELP)
@click.argument("target", required=False, default="default")
@click.option("-l", "--list-targets", is_flag=True, is_eager=True,
              expose_value=False, callback=list_targets_callback,
              help="List available project names defined in your config")
@click.option("-d", "--debug", help="debug mode", is_flag=True,
              callback=utils.deprecate_debug_parameter)
@click.option("-v", "--verbose", is_flag=True,
              help="stream deploy logs to console")
@click.option("-V", "--version", help="release version")
@click.option("--username", help="docker registry name")
@click.option("--password", help="docker registry password")
@click.option("--email", help="docker registry email")
@click.option("--apikey", help="SH apikey to use built-in registry")
@click.option("--insecure", is_flag=True, help="use insecure registry")
@click.option("--async", "async_", is_flag=True, help="[DEPRECATED] enable asynchronous mode",
              callback=utils.deprecate_async_parameter)
def cli(target, debug, verbose, version, username, password, email,
        apikey, insecure, async_):
    deploy_cmd(target, version, username, password, email,
               apikey, insecure, async_)


def deploy_cmd(target, version, username, password, email,
               apikey, insecure, async_):
    config = load_shub_config()
    target_conf = config.get_target_conf(target)
    endpoint, target_apikey = target_conf.endpoint, target_conf.apikey
    image = config.get_image(target)
    version = version or config.get_version()
    image_name = utils.format_image_name(image, version)
    username, password = utils.get_credentials(
        username=username, password=password, insecure=insecure,
        apikey=apikey, target_apikey=target_apikey)

    apikey = apikey or target_apikey
    params = _prepare_deploy_params(
        target_conf.project_id, version, image_name, endpoint, apikey,
        username, password, email)

    click.echo(f"Deploying {image_name}")
    utils.debug_log(f'Deploy parameters: {params}')
    req = requests.post(
        urljoin(endpoint, '/api/releases/deploy.json'),
        data=params,
        auth=(apikey, ''),
        timeout=300,
        allow_redirects=False
    )
    if req.status_code == 400:
        reason = req.json().get('non_field_errors')
        raise ShubException('\n'.join(reason) if reason else req.text)
    req.raise_for_status()
    status_url = req.headers['location']
    status_id = utils.store_status_url(
        status_url, limit=STORE_N_LAST_STATUS_URLS)
    click.echo(
        "You can check deploy results later with "
        "'shub image check --id {}'.".format(status_id))
    if async_:
        return
    if utils.is_verbose():
        deploy_progress_cls = _LoggedDeployProgress
    else:
        deploy_progress_cls = _DeployProgress
    events = _convert_status_requests_to_events(status_url)
    deploy_progress = deploy_progress_cls(events)
    deploy_progress.show()


def _convert_status_requests_to_events(status_url):
    """Convert a sequence of deploy status requests to an iterator.

    Current logic yields only distinct events and stops execution
    when event status is a final one.
    """
    previous_event = None
    while True:
        event = _check_status_url(status_url)
        # no need to handle already seen events
        if previous_event != event:
            yield event
            if event['status'] not in SYNC_DEPLOY_WAIT_STATUSES:
                break
            previous_event = event
        time.sleep(SYNC_DEPLOY_REFRESH_TIMEOUT)


class _LoggedDeployProgress(utils.BaseProgress):
    """Visualize deploy progress in verbose mode."""

    def show(self):
        click.echo("Deploy results:")
        super().show()

    def handle_event(self, event):
        click.echo(event)


class _DeployProgress(utils.BaseProgress):
    """Visualize deploy progress in non-verbose mode.

    Uses a progress bar to track total progress.
    """
    def __init__(self, events):
        super().__init__(events)
        self.progress_bar = self._create_progress_bar()
        self.result_event = None

    def show(self):
        super().show()
        # it's possible that release process finishes instantly without
        # providing enough information to fill progress bar completely
        if (
            self.result_event
            and self.result_event['status'] == 'ok'
            and self.progress_bar.n != self.progress_bar.total
        ):
            delta = max(self.progress_bar.total - self.progress_bar.n, 0)
            self.progress_bar.update(delta)
        self.progress_bar.close()
        # last event with non-waiting status contains successful result or
        # error result from the service with error details
        if self.result_event:
            click.echo(f"Deploy results:\n{self.result_event}")

    def handle_event(self, event):
        if 'progress' in event and 'total' in event:
            self.progress_bar.total = event['total']
            self.progress_bar.update(max(event['progress'] - self.progress_bar.n, 0))
        elif event['status'] not in SYNC_DEPLOY_WAIT_STATUSES:
            self.result_event = event

    def _create_progress_bar(self):
        return utils.create_progress_bar(
            total=DEFAULT_TOTAL_PROGRESS,
            desc='Progress',
            # don't need rate here, let's simplify the bar
            bar_format='{l_bar}{bar}| {n_fmt}/{total_fmt}'
        )


def _retry_on_requests_error(exception):
    return isinstance(exception, CHECK_RETRY_EXCEPTIONS)


@retry(retry_on_exception=_retry_on_requests_error,
       stop_max_attempt_number=CHECK_RETRY_ATTEMPTS,
       wait_exponential_multiplier=CHECK_RETRY_EXP_MULTIPLIER,
       wait_exponential_max=CHECK_RETRY_EXP_MAX)
def _check_status_url(status_url):
    status_req = requests.get(status_url, timeout=300)
    status_req.raise_for_status()
    return status_req.json()


def _prepare_deploy_params(project, version, image_name, endpoint, apikey,
                           username, password, email):
    # Reusing shub.image.list logic to get spiders list
    metadata = list_mod.list_cmd(image_name, project, endpoint, apikey)
    if not metadata.get('scripts'):
        metadata['scripts'] = _extract_scripts_from_project()
    params = {
        'project': project,
        'version': version,
        'image_url': image_name,
        'project_type': metadata['project_type'],
    }
    if metadata.get('spiders'):
        params['spiders'] = ','.join(metadata['spiders'])
    if metadata.get('scripts'):
        params['scripts'] = ','.join(metadata['scripts'])
    if not username:
        params['pull_insecure_registry'] = True
    else:
        params['pull_auth_config'] = json.dumps(
            {'username': username,
             'password': password,
             'email': email}, sort_keys=True)
    return params


def _extract_scripts_from_project(setup_filename='setup.py'):
    """Parse setup.py and return scripts"""
    if not os.path.isfile(setup_filename):
        return []
    mock_setup = textwrap.dedent('''\
    def setup(*args, **kwargs):
        __setup_calls__.append((args, kwargs))
    ''')
    parsed_mock_setup = ast.parse(mock_setup, filename=setup_filename)
    with open(setup_filename) as setup_file:
        parsed = ast.parse(setup_file.read())
        for index, node in enumerate(parsed.body[:]):
            if (not isinstance(node, ast.Expr) or
                    not isinstance(node.value, ast.Call) or
                    node.value.func.id != 'setup'):
                continue
            parsed.body[index:index] = parsed_mock_setup.body
            break
    fixed = ast.fix_missing_locations(parsed)
    codeobj = compile(fixed, setup_filename, 'exec')
    local_vars = {}
    global_vars = {'__setup_calls__': []}
    exec(codeobj, global_vars, local_vars)
    _, kwargs = global_vars['__setup_calls__'][0]
    return [os.path.basename(f) for f in kwargs.get('scripts', [])]



import os
import unittest
from unittest import mock

import yaml
from click.testing import CliRunner
from yaml import CLoader as Loader

from shub.migrate_eggs import main
from shub.config import Target


class MigrateEggsTest(unittest.TestCase):
    REQ_LIST = [
        'boto==2.38.0',
        'dateparser==0.3.3',
        'decorator==4.0.10',
        'dicttoxml==1.6.6',
        'httpretty==0.8.0',
        'hubstorage==0.16.1',
        'jdatetime==1.7.2',
        'six==1.9.0', 'spur==0.3.15',
        'SQLAlchemy==1.0.5',
        'sqlitedict==1.3.0',
        'urllib3==1.11',
        'wheel==0.24.0',
        'wsgiref==0.1.2',
    ]

    def run(self, *a, **kw):
        runner = CliRunner()
        with runner.isolated_filesystem():
            super().run(*a, **kw)

    def setUp(self):
        self.clickm = mock.patch('shub.migrate_eggs.click').start()
        gtc = mock.patch('shub.migrate_eggs.get_target_conf').start()
        self.requestsm = mock.patch('shub.migrate_eggs.requests').start()

        self.curr_dir = os.path.dirname(os.path.realpath(__file__))

        with open('./scrapinghub.yml', 'w') as f:
            f.write('')

        gtc.return_value = Target(
            project_id=123,
            endpoint='endpoint1',
            apikey='apikey1',
            stack='',
            image='',
            requirements_file='',
            version='',
            eggs=[],
        )

        self.addCleanup(mock.patch.stopall)

    def walksorted(self):
        return [
            (sorted(dirs), sorted(files))
            for _, dirs, files in os.walk('.')
        ]

    def _assert_requirements_content(self):
        with open('./requirements.txt') as f:
            content = f.read()
            self.assertIn('DISABLE_DASH_EGGS', content)
            requirements = [line for line in content.split('\n') if '==' in line]
            self.assertListEqual(requirements, self.REQ_LIST)

    def test_full(self):
        migrate_zip = os.path.join(self.curr_dir, 'samples/migrate-eggs.zip')
        with open(migrate_zip, 'rb') as f:
            self.requestsm.get().content = f.read()

        main('default')
        self.clickm.confirm.assert_called_with(
            'Eggs will be stored in ./eggs, are you sure ? '
        )

        files = self.walksorted()

        self.assertEqual(
            files[0],
            (['eggs'], ['requirements.txt', 'scrapinghub.yml']),
        )
        self.assertEqual(
            files[1],
            ([], ['1.egg', '2.egg', '3.egg'])
        )

        with open('./scrapinghub.yml') as f:
            abc = yaml.load(f, Loader=Loader)
            eggs = abc['requirements'].pop('eggs')
            eggs = [e.replace('\\', '/') for e in eggs]
            self.assertEqual(
                eggs,
                [
                    './eggs/1.egg',
                    './eggs/2.egg',
                    './eggs/3.egg',
                ],
            )
            self.assertDictEqual(
                abc,
                {
                    'requirements': {
                        'file': './requirements.txt'
                    },
                }
            )

        self._assert_requirements_content()

        for i in range(1, 4):
            i = str(i)
            with open('./eggs/%s.egg' % i) as f:
                self.assertEqual(f.read().strip(), i)

    def test_no_eggs(self):
        file_ = 'samples/migrate-eggs-no-eggs.zip'
        migrate_zip = os.path.join(self.curr_dir, file_)
        with open(migrate_zip, 'rb') as f:
            self.requestsm.get().content = f.read()

        main('default')
        self.assertFalse(self.clickm.confirm.called)

        files = self.walksorted()
        self.assertListEqual(
            files,
            [([], ['requirements.txt', 'scrapinghub.yml'])]
        )

        with open('./scrapinghub.yml') as f:
            abc = yaml.load(f, Loader=Loader)
            self.assertDictEqual(
                abc,
                {
                    'requirements': {
                        'file': './requirements.txt'
                    },
                }
            )

        self._assert_requirements_content()

    def test_override_reqs_file(self):
        file_ = 'samples/migrate-eggs-no-eggs.zip'
        migrate_zip = os.path.join(self.curr_dir, file_)
        with open(migrate_zip, 'rb') as f:
            self.requestsm.get().content = f.read()
        with open('./requirements.txt', 'w') as f:
            f.write('smth==1.2.3')

        self.clickm.confirm.return_value = False
        main('default')
        self.clickm.confirm.assert_called_with(
            'requirements.txt already exists, are you sure to override it ?'
        )

        files = self.walksorted()
        self.assertListEqual(
            files,
            [([], ['requirements.txt', 'scrapinghub.yml'])]
        )

        with open('./scrapinghub.yml') as f:
            self.assertEqual(f.read(), '')

        with open('./requirements.txt') as f:
            content = f.read()
            self.assertEqual(content, 'smth==1.2.3')

        self.clickm.reset_mock()
        self.clickm.confirm.return_value = True
        main('default')

        self.clickm.confirm.assert_called_with(
            'requirements.txt already exists, are you sure to override it ?'
        )

        files = self.walksorted()
        self.assertListEqual(
            files,
            [([], ['requirements.txt', 'scrapinghub.yml'])]
        )

        with open('./scrapinghub.yml') as f:
            abc = yaml.load(f, Loader=Loader)
            self.assertDictEqual(
                abc,
                {
                    'requirements': {
                        'file': './requirements.txt'
                    },
                }
            )

        self._assert_requirements_content()



import textwrap
import unittest
from unittest import mock

from click.testing import CliRunner

from shub import config, logout


@mock.patch('shub.config.GLOBAL_SCRAPINGHUB_YML_PATH', new='.scrapinghub.yml')
@mock.patch('shub.config.NETRC_PATH', new='.netrc')
@mock.patch('shub.logout.GLOBAL_SCRAPINGHUB_YML_PATH', new='.scrapinghub.yml')
class LogoutTestCase(unittest.TestCase):

    def setUp(self):
        self.runner = CliRunner()

    def test_remove_key(self):
        GLOBAL_SH_YML = textwrap.dedent("""
            apikeys:
                default: LOGGED_IN_KEY
        """)
        with self.runner.isolated_filesystem():
            with open('.scrapinghub.yml', 'w') as f:
                f.write(GLOBAL_SH_YML)
            conf = config.load_shub_config()
            self.assertIn('default', conf.apikeys)
            self.runner.invoke(logout.cli)
            conf = config.load_shub_config()
            self.assertNotIn('default', conf.apikeys)

    @mock.patch('shub.logout.update_yaml_dict')
    def test_fail_on_not_logged_in(self, mock_uyd):
        with self.runner.isolated_filesystem():
            self.runner.invoke(logout.cli)
            self.assertFalse(mock_uyd.called)



from setuptools import setup, find_packages


setup(
    name='shub',
    version='2.16.0',
    packages=find_packages(exclude=('tests', 'tests.*')),
    url="https://shub.readthedocs.io/en/stable/",
    description='Scrapinghub Command Line Client',
    long_description=open('README.rst').read(),
    author='Scrapinghub',
    author_email='info@scrapinghub.com',
    maintainer='Scrapinghub',
    maintainer_email='info@scrapinghub.com',
    license='BSD',
    entry_points={
        'console_scripts': ['shub = shub.tool:cli']
    },
    include_package_data=True,
    zip_safe=False,
    python_requires='>=3.9',
    install_requires=[
        'click',
        'docker',
        'importlib-metadata; python_version < "3.10"',
        'packaging',
        'pip',
        'PyYAML',
        'retrying',
        'requests',
        'scrapinghub>=2.3.1',
        'setuptools',
        'tqdm==4.55.1',
        'toml',
    ],
    classifiers=[
        'Development Status :: 5 - Production/Stable',
        'Intended Audience :: Developers',
        'Natural Language :: English',
        'License :: OSI Approved :: BSD License',
        'Programming Language :: Python',
        'Programming Language :: Python :: 3.9',
        'Programming Language :: Python :: 3.10',
        'Programming Language :: Python :: 3.11',
        'Programming Language :: Python :: 3.12',
        'Programming Language :: Python :: 3.13',
        'Programming Language :: Python :: 3.14',
        'Operating System :: OS Independent',
        'Environment :: Console',
        'Topic :: Internet :: WWW/HTTP',
    ],
)



import os
import zipfile
from unittest import mock

import pytest
import requests
import yaml
from click.testing import CliRunner

from shub.bootstrap import cli, EXAMPLE_REPO, list_projects, unzip_project
from shub.exceptions import (
    BadParameterException, NotFoundException, RemoteErrorException)


BOOTSTRAP_PROJECTS = """
projA:
    path: projects_dir/projA
    description: PROJECT_A_DESC

projB:
    path: projects_dir/projB
    description: PROJECT_B_DESC
"""


REPO_ZIP_PATH = os.path.join(os.path.dirname(__file__), 'samples',
                             'custom-images-examples-master.zip')


@pytest.fixture
def requests_get_mock():
    with mock.patch('shub.bootstrap.requests.get') as m:
        yield m


@pytest.fixture
def github_responses(requests_get_mock):
    requests_get_mock.return_value.text = BOOTSTRAP_PROJECTS
    with open(REPO_ZIP_PATH, 'rb') as f:
        requests_get_mock.return_value.content = f.read()


def test_list_projects(capsys):
    projects = yaml.safe_load(BOOTSTRAP_PROJECTS)
    list_projects(projects)
    out, _ = capsys.readouterr()
    assert 'projA' in out
    assert 'PROJECT_A_DESC' in out
    assert 'projB' in out
    assert 'PROJECT_B_DESC' in out
    assert 'projects_dir' not in out


def test_unzip_project(tempdir):
    target_dir = str(tempdir.join('projA'))
    project = {'path': 'projects_dir/projA'}
    repo_zip = zipfile.ZipFile(REPO_ZIP_PATH)
    assert not os.path.exists(target_dir)
    unzip_project(repo_zip, project, target_dir)
    assert os.path.exists(target_dir)
    assert os.path.isfile(os.path.join(target_dir, 'a_file'))
    assert os.path.isdir(os.path.join(target_dir, 'a_dir'))
    assert os.path.isfile(os.path.join(target_dir, 'a_dir', 'a_dir_file'))


@pytest.mark.usefixtures('github_responses')
def test_cli_lists_projects():
    result = CliRunner().invoke(cli, ['-l'])
    assert result.exit_code == 0
    assert 'projA' in result.output
    assert 'PROJECT_A_DESC' in result.output


@pytest.mark.usefixtures('github_responses')
def test_cli_clones_project_into_default_dir(tempdir):
    target_dir = str(tempdir.join('projA'))
    assert not os.path.exists(target_dir)
    result = CliRunner().invoke(cli, ['projA'])
    assert result.exit_code == 0
    assert os.path.isdir(target_dir)
    assert os.path.isfile(os.path.join(target_dir, 'a_file'))


@pytest.mark.usefixtures('github_responses')
def test_cli_clones_project_into_target_dir(tempdir):
    target_dir = str(tempdir.join('target_dir'))
    assert not os.path.exists(target_dir)
    result = CliRunner().invoke(cli, ['projA', 'target_dir'])
    assert result.exit_code == 0
    assert os.path.isdir(target_dir)
    assert os.path.isfile(os.path.join(target_dir, 'a_file'))


def test_cli_fails_on_existing_target_dir(tempdir):
    os.mkdir('target_dir')
    result = CliRunner().invoke(cli, ['some_project', 'target_dir'])
    assert result.exit_code == BadParameterException.exit_code
    assert "exists" in result.output


@pytest.mark.usefixtures('github_responses')
def test_cli_fails_on_unknown_project():
    result = CliRunner().invoke(cli, ['nonexistent'])
    assert result.exit_code == NotFoundException.exit_code
    assert "shub bootstrap -l" in result.output


def test_cli_links_to_repo_on_http_error(requests_get_mock):
    requests_get_mock.return_value.raise_for_status.side_effect = (
        requests.HTTPError)
    result = CliRunner().invoke(cli, ['some_project'])
    assert result.exit_code == RemoteErrorException.exit_code
    assert EXAMPLE_REPO in result.output



import unittest
from collections import namedtuple
from unittest import mock

from click.testing import CliRunner

from shub import cancel
from shub.exceptions import (
    BadParameterException,
    ShubException,
)

from .utils import AssertInvokeRaisesMixin, mock_conf


class CancelTest(AssertInvokeRaisesMixin, unittest.TestCase):

    def setUp(self):
        self.runner = CliRunner()
        self.conf = mock_conf(self)

    @mock.patch('shub.cancel.get_scrapinghub_client_from_config')
    def test_simple_cancel_call(self, mock_client):
        client = mock_client.return_value
        mock_proj = client.get_project.return_value
        mock_proj.jobs.cancel.return_value = {'count': 2}

        result = self.runner.invoke(
            cancel.cli, ('123456', '1/1', '1/2',), input='y\n'
        )

        self.assertTrue("{'count': 2}" in result.output)
        self.assertEqual(0, result.exit_code)
        self.assertEqual(mock_proj.jobs.cancel.call_count, 1)
        mock_proj.jobs.cancel.assert_called_with(
            keys=['123456/1/1', '123456/1/2']
        )

    @mock.patch('shub.cancel.get_target_conf')
    @mock.patch('shub.cancel.get_scrapinghub_client_from_config')
    def test_cancel_default_project(self, mock_client, targetconf):
        client = mock_client.return_value
        mock_proj = client.get_project.return_value
        mock_proj.jobs.cancel.return_value = {'count': 2}

        Target = namedtuple('Target', 'project_id')
        targetconf.return_value = Target(project_id='123456')

        result = self.runner.invoke(
            cancel.cli, ('1/1', '1/2',), input='y\n'
        )

        self.assertTrue("{'count': 2}" in result.output)
        self.assertEqual(0, result.exit_code)
        self.assertEqual(mock_proj.jobs.cancel.call_count, 1)
        mock_proj.jobs.cancel.assert_called_with(
            keys=['123456/1/1', '123456/1/2']
        )

    @mock.patch('shub.cancel.get_scrapinghub_client_from_config')
    def test_invalid_job_key(self, mock_client):
        self.assertInvokeRaises(
            SystemExit,
            cancel.cli,
            ('123456', '1/1', '1',),
            input='y\n'
        )

        self.assertInvokeRaises(
            SystemExit,
            cancel.cli,
            ('123456', '1/abc', '1',),
            input='y\n'
        )

    @mock.patch('shub.cancel.get_scrapinghub_client_from_config')
    def test_cancel_failure(self, mock_client):
        client = mock_client.return_value
        mock_proj = client.get_project.return_value
        mock_proj.jobs.cancel.side_effect = ValueError('Error msg')

        self.assertInvokeRaises(
            ShubException,
            cancel.cli,
            ('123456', '1/1', '1/2',),
            input='y\n',
        )

    @mock.patch('shub.cancel.get_scrapinghub_client_from_config')
    def test_cancel_abort(self, mock_client):
        client = mock_client.return_value
        client.get_project.return_value

        result = self.runner.invoke(
            cancel.cli, ('123456', '1/1', '1/2',), input='N\n',
        )
        self.assertTrue('Aborted!' in result.output)

    def test_validate_job_key(self):
        with self.assertRaises(BadParameterException):
            cancel.validate_job_key('123456', '1')

        with self.assertRaises(BadParameterException):
            cancel.validate_job_key('123456', '1/abc')

        with self.assertRaises(BadParameterException):
            cancel.validate_job_key('123456', '')



import netrc
import os
import warnings
from collections import namedtuple
from urllib.parse import urlparse, urlunparse

import click
import yaml

from shub import DOCS_LINK, CONFIG_DOCS_LINK
from shub.exceptions import (BadParameterException, BadConfigException,
                             ConfigParseException, MissingAuthException,
                             NotFoundException, ShubDeprecationWarning,
                             print_warning)
from shub.utils import (closest_file, get_scrapycfg_targets, get_sources,
                        pwd_hg_version, pwd_git_version, pwd_version,
                        update_yaml_dict)

APIKEY_SHOW_N_CHARS = 6
SH_IMAGES_REGISTRY = 'images.scrapinghub.com'
SH_IMAGES_REPOSITORY = SH_IMAGES_REGISTRY + '/project/{project}'
GLOBAL_SCRAPINGHUB_YML_PATH = os.path.expanduser(
    os.environ.get('SHUB_GLOBAL_CONFIG', '~/.scrapinghub.yml')
)
NETRC_PATH = os.path.expanduser('~/_netrc' if os.name == 'nt' else '~/.netrc')


class ShubConfig:

    DEFAULT_ENDPOINT = 'https://app.zyte.com/api/'

    # Dictionary option name: Shortcut to set 'default' key
    SHORTCUTS = {
        'projects': 'project',
        'endpoints': 'endpoint',
        'apikeys': 'apikey',
        'stacks': 'stack',
        'images': 'image',
    }

    def __init__(self):
        self.projects = {}
        self.endpoints = {
            'default': self.DEFAULT_ENDPOINT,
        }
        self.apikeys = {}
        self.version = 'AUTO'
        self.stacks = {}
        self.requirements_file = None
        self.eggs = []
        self.images = {}

    def _check_endpoints(self):
        """Check the endpoints. Send warnings if necessary."""
        for endpoint, url in self.endpoints.items():
            parsed = urlparse(url)
            if parsed.netloc == 'staging.scrapinghub.com':
                self.endpoints[endpoint] = urlunparse(
                    parsed._replace(netloc='app.zyte.com')
                )
                click.echo(
                    'WARNING: Endpoint "%s" is still using %s which has been '
                    'obsoleted. shub has updated it to app.zyte.com '
                    'for this time only. Please update your configuration.' % (
                        endpoint, parsed.netloc,
                    ),
                    err=True
                )
            if parsed.scheme == 'http':
                print_warning(
                    'Endpoint "%s" is still using HTTP. '
                    'Please change it to HTTPS if possible.' % endpoint
                )

    def load(self, stream):
        """Load Scrapinghub configuration from stream."""
        # flag to mark if images/default was used in the config file
        check_default_image_scope = False
        try:
            yaml_cfg = yaml.safe_load(stream)
            if not yaml_cfg:
                return
            for option, shortcut in self.SHORTCUTS.items():
                option_conf = getattr(self, option)
                yaml_option_conf = yaml_cfg.get(option, {})
                option_conf.update(yaml_option_conf)
                if option == 'images' and yaml_option_conf:
                    print_warning(
                        "Images section is deprecated, please replace it with "
                        "global `image` setting or define `image` setting for "
                        "the project.\n  Check for additional details in {}."
                        .format(CONFIG_DOCS_LINK),
                        category=ShubDeprecationWarning
                    )
                    if 'default' in yaml_option_conf:
                        check_default_image_scope = True
                if shortcut in yaml_cfg:
                    # We explicitly check yaml_option_conf and not option_conf.
                    # It is okay to set conflicting defaults if they are in
                    # different files (b/c then one of these will have
                    # priority)
                    if 'default' in yaml_option_conf:
                        raise BadConfigException(
                            "You cannot specify both '%s' and a 'default' key "
                            "for '%s' in the same file" % (shortcut,  option))
                    option_conf['default'] = yaml_cfg[shortcut]
            self.version = yaml_cfg.get('version', self.version)
            self.requirements_file = yaml_cfg.get('requirements_file',
                                                  self.requirements_file)
            self.requirements_file = yaml_cfg.get('requirements', {}).get(
                'file', self.requirements_file)
            self.eggs = yaml_cfg.get('requirements', {}).get('eggs', self.eggs)
        except (yaml.YAMLError, AttributeError):
            # AttributeError: stream is valid YAML but not dictionary-like
            raise ConfigParseException
        # fail if `projects` section has keys not found in `images`
        if (check_default_image_scope and
                not set(self.projects).issubset(set(self.images))):
            raise BadConfigException(
                    "Found ambigious configuration: default image has global "
                    "scope now, but some projects were not using custom "
                    "images and can be broken now. Please fix your config by "
                    "replacing `images` section with `image` settings. Check "
                    "for additional details in {}".format(CONFIG_DOCS_LINK)
                )
        self._check_endpoints()

    def load_file(self, filename):
        """Load Scrapinghub configuration from YAML file. """
        try:
            with open(filename) as f:
                self.load(f)
        except ConfigParseException:
            raise ConfigParseException(
                "Unable to parse configuration file %s. Maybe a missing "
                "colon?" % filename
            )

    def _load_scrapycfg_target(self, tname, t):
        default_endpoint = ('url' not in t or
                            t['url'] == self.endpoints['default'])
        default_user = ('username' not in t or
                        t['username'] == self.apikeys.get('default'))
        if 'project' in t:
            if tname == 'default' or (default_endpoint and default_user):
                self.projects[tname] = t['project']
            elif default_endpoint and not default_user:
                self.projects[tname] = {
                    'id': t['project'], 'apikey': tname,
                }
            else:
                self.projects[tname] = tname + '/' + t['project']
        if not default_endpoint:
            self.endpoints[tname] = t['url']
        if not default_user or (not default_endpoint and 'username' in t):
            self.apikeys[tname] = t['username']
        if 'version' in t:
            self.version = t['version']

    def load_scrapycfg(self, sources):
        """Load configuration from a list of scrapy.cfg-like sources."""
        targets = get_scrapycfg_targets(sources)
        self._load_scrapycfg_target('default', targets['default'])
        del targets['default']
        for tname, t in targets.items():
            self._load_scrapycfg_target(tname, t)
        self._check_endpoints()

    def save(self, path=None, options=None):
        def _project_id_as_int(project):
            """Copy project and return it with the ID casted to int to make
            sure it is exported as "123" and not "'123'"
            """
            try:
                if isinstance(project, dict):
                    project = project.copy()
                    project['id'] = int(project['id'])
                else:
                    project = int(project)
            except ValueError:
                # Happens when project ID contains endpoint (e.g. vagrant/123),
                # in that case it will be exported without quotation marks
                # anyway
                pass
            return project

        with update_yaml_dict(path) as yml:
            options = options or list(self.SHORTCUTS)
            for option in options:
                shortcut = self.SHORTCUTS[option]
                conf = getattr(self, option)
                if option == 'endpoints' and conf == ShubConfig().endpoints:
                    # Don't write default endpoint
                    continue
                elif option == 'projects':
                    conf = {k: _project_id_as_int(v) for k, v in conf.items()}
                if list(conf.keys()) == ['default']:
                    yml[shortcut] = conf['default']
                    yml.pop(option, None)
                else:
                    yml[option] = conf
                    yml.pop(shortcut, None)
            if self.version != 'AUTO':
                yml['version'] = self.version
            if self.eggs:
                yml.setdefault('requirements', {})['eggs'] = self.eggs
            if self.requirements_file:
                yml.setdefault('requirements', {})['file'] = (
                    self.requirements_file)

    @property
    def normalized_projects(self):
        """
        Return a copy of ``self.projects`` where all values are dictionaries
        that have at least the keys ``id``, ``endpoint``, and ``apikey``.
        """
        projects = self.projects.copy()
        for target, proj in list(projects.items()):
            if not isinstance(proj, dict):
                proj = {'id': proj}
                projects[target] = proj
            elif 'id' not in proj:
                raise BadConfigException("Please define an ID for project "
                                         "\"%s\"" % target)
            try:
                proj['endpoint'], proj['id'] = proj['id'].split('/')
            except (ValueError, AttributeError):
                proj.setdefault('endpoint', 'default')
            proj.setdefault('apikey', proj['endpoint'])
            try:
                proj['id'] = int(proj['id'])
            except ValueError:
                raise BadConfigException(
                    "\"%s\" is not a valid Scrapinghub project ID. Please "
                    "check your scrapinghub.yml" % proj['id']
                )
        return projects

    def get_project(self, project):
        """
        Given a project alias or a canonical project ID, return the
        corresponding normalized configuration dictionary from
        ``self.projects``.
        """
        if project in self.projects:
            return self.normalized_projects[project]
        try:
            endpoint, proj_id = project.split('/')
        except (ValueError, AttributeError):
            endpoint, proj_id = 'default', project
        try:
            proj_id = int(proj_id)
        except ValueError:
            if project == 'default':
                msg = ("Please specify target or configure a default target "
                       "in scrapinghub.yml.")
            else:
                msg = ("Could not find target \"%s\". Please define it in "
                       "your scrapinghub.yml or supply a numerical project ID."
                       "" % project)
            raise BadParameterException(msg, param_hint='target')
        for proj in self.normalized_projects.values():
            if proj['id'] == proj_id and proj['endpoint'] == endpoint:
                return proj
        else:
            return {'id': proj_id, 'endpoint': endpoint, 'apikey': endpoint}

    def get_version(self):
        if not self.version or self.version == 'AUTO':
            return pwd_version()
        elif self.version == 'GIT':
            return pwd_git_version()
        elif self.version == 'HG':
            return pwd_hg_version()
        elif self.version:
            return str(self.version)

    def get_target_conf(self, target, auth_required=True):
        proj = self.get_project(target)
        if proj['endpoint'] not in self.endpoints:
            raise NotFoundException("Could not find endpoint %s. Please "
                                    "define it in your scrapinghub.yml."
                                    "" % proj['endpoint'])
        try:
            apikey = str(self.apikeys[proj['apikey']])
        except KeyError:
            if auth_required:
                msg = None
                if proj['endpoint'] != 'default':
                    msg = ("Could not find API key for endpoint %s."
                           "" % proj['endpoint'])
                raise MissingAuthException(msg)
            apikey = None
        proj_requirements = proj.get('requirements', {})
        requirements = proj_requirements.get('file', self.requirements_file)
        eggs = proj_requirements.get('eggs', self.eggs)
        return Target(
            project_id=proj['id'],
            endpoint=self.endpoints[proj['endpoint']],
            apikey=apikey,
            stack=(self.stacks.get(proj['stack'], proj['stack'])
                   if 'stack' in proj else self.stacks.get('default')),
            image=self._select_image_for_project(target, proj),
            requirements_file=requirements,
            version=self.get_version(),
            eggs=eggs,
        )

    def _select_image_for_project(self, target, project):
        """Helper to select image for a project (or its target).

        The select logic is the following:
        - image defined per project has highest priority,
        - images section is marked as deprecated, but still in force:
          if target is defined in images - use a corresponding image,
        - if image is not defined, but there's a default image set by
          image or images/default settings - use it.

        The function responds with a custom image name string
        (or None/False meaning regular stack-based deploy).
        """
        image = project.get('image', self.images.get(
            target, self.images.get('default')))
        # aliases to use internal scrapinghub registry as image storage
        if image is True or image == 'scrapinghub':
            image = SH_IMAGES_REPOSITORY.format(project=project['id'])
        return image

    def get_target(self, target, auth_required=True):
        """Return (project_id, endpoint, apikey) for given target."""
        warnings.warn("get_target is deprecated, use get_target_conf instead")
        targetconf = self.get_target_conf(target, auth_required=auth_required)
        return (
            targetconf.project_id,
            targetconf.endpoint,
            targetconf.apikey
        )

    def get_project_id(self, target):
        return self.get_target_conf(target, auth_required=False).project_id

    def get_endpoint(self, target):
        return self.get_target_conf(target, auth_required=False).endpoint

    def get_apikey(self, target, required=True):
        apikey = self.get_target_conf(target, auth_required=required).apikey
        return getattr(apikey, 'value', apikey)

    def get_image(self, target):
        """Return image for a given target."""
        target_conf = self.get_target_conf(target, auth_required=False)
        project, image = target_conf.project_id, target_conf.image
        if image is None:
            raise NotFoundException(
                "Could not find image for project '{}'. Please define it "
                "in your scrapinghub.yml.".format(target))
        elif image is False:
            raise BadConfigException(
                "Using custom images is disabled for the project '{}'. "
                "Please enable it in your scrapinghub.yml.".format(target))
        elif target_conf.stack:
            raise BadConfigException(
                "Ambiguous configuration: There is both a custom image and a "
                "stack configured for project '{}'. Please see {} for "
                "information on how to configure both custom image-based and "
                "stack-based projects.".format(target, CONFIG_DOCS_LINK))
        default_image = SH_IMAGES_REPOSITORY.format(project=project)
        if image.startswith(SH_IMAGES_REGISTRY) and image != default_image:
            raise BadConfigException(
                "Found wrong SH repository for project '{}': expected {}.\n  "
                "Please use aliases `True` or `scrapinghub` to fix it in your "
                "config.".format(target, default_image))
        return image


_Target = namedtuple('Target', ['project_id', 'endpoint', 'apikey', 'stack',
                                'image', 'requirements_file', 'version',
                                'eggs'])


class APIkey(str):

    def __new__(cls, *args, **kwargs):
        cls._inst = super().__new__(cls, *args, **kwargs)
        return cls._inst

    def __init__(self, value=None):
        self.value = value

    def __repr__(self):
        if not self.value:
            return ''
        visible_chars = APIKEY_SHOW_N_CHARS
        return (self.value[:visible_chars] +
                'X' * max(len(self.value) - visible_chars, 0))


class Target(_Target):

    def __new__(cls, project_id, endpoint, apikey, *args, **kwargs):
        cls._inst = super().__new__(cls, project_id, endpoint,
                                    APIkey(apikey), *args, **kwargs)
        return cls._inst


MIGRATION_BANNER = """
-------------------------------------------------------------------------------
Welcome to shub version 2!

This release contains major updates to how shub is configured, as well as
updates to the commands and shub's look & feel.

Run 'shub' to get an overview over all available commands, and
'shub command --help' to get detailed help on a command. Definitely try the
new 'shub items -f [JOBID]' to see items live as they are being scraped!

From now on, shub configuration should be done in a file called
'scrapinghub.yml', living next to the previously used 'scrapy.cfg' in your
Scrapy project directory. Global configuration, for example API keys, should be
done in a file called '.scrapinghub.yml' in your home directory.

But no worries, shub has automatically migrated your global settings to
~/.scrapinghub.yml, and will also automatically migrate your project settings
when you run a command within a Scrapy project.

Visit {docs_link} for more information on the new configuration format and
its benefits.

Happy scraping!
-------------------------------------------------------------------------------
""".format(docs_link=DOCS_LINK)


def _migrate_to_global_scrapinghub_yml():
    conf = ShubConfig()
    conf.load_scrapycfg(get_sources(use_closest=False))
    try:
        info = netrc.netrc(NETRC_PATH)
        netrc_key, _, _ = info.authenticators("scrapinghub.com")
    except (OSError, TypeError):
        netrc_key = None
    if netrc_key:
        conf.apikeys['default'] = netrc_key
    conf.save(GLOBAL_SCRAPINGHUB_YML_PATH)
    default_conf = ShubConfig()
    migrated_data = any(getattr(conf, attr) != getattr(default_conf, attr)
                        for attr in ('projects', 'endpoints', 'apikeys',
                                     'version'))
    if migrated_data:
        click.echo(MIGRATION_BANNER, err=True)


PROJECT_MIGRATION_OK_BANNER = """
INFO: Your deploy configuration has been migrated to scrapinghub.yml.
shub will no longer read from scrapy.cfg (but Scrapy will, so don't delete it).
Visit {docs_link} for more information.
""".format(docs_link=DOCS_LINK)


PROJECT_MIGRATION_FAILED_BANNER = """
WARNING: shub failed to convert your scrapy.cfg to scrapinghub.yml. Please
visit {docs_link} for help on how to use the new configuration format. We
would be grateful if you could also file a bug report at
https://github.com/scrapinghub/shub/issues

For now, shub fell back to reading from scrapy.cfg, everything should work as
expected.
""".format(docs_link=DOCS_LINK)


def _migrate_and_load_scrapy_cfg(conf):
    # Load from closest scrapy.cfg
    closest_scrapycfg = closest_file('scrapy.cfg')
    if not closest_scrapycfg:
        return
    targets = get_scrapycfg_targets([closest_scrapycfg])
    if targets == get_scrapycfg_targets():
        # No deploy configuration in scrapy.cfg
        return
    conf.load_scrapycfg([closest_scrapycfg])
    # Migrate to scrapinghub.yml
    closest_sh_yml = os.path.join(os.path.dirname(closest_scrapycfg),
                                  'scrapinghub.yml')
    temp_conf = ShubConfig()
    temp_conf.load_scrapycfg([closest_scrapycfg])
    try:
        temp_conf.save(closest_sh_yml)
    except Exception:
        click.echo(PROJECT_MIGRATION_FAILED_BANNER, err=True)
    else:
        click.echo(PROJECT_MIGRATION_OK_BANNER, err=True)


def load_shub_config(load_global=True, load_local=True, load_env=True):
    """
    Return a ShubConfig instance with ~/.scrapinghub.yml and the closest
    scrapinghub.yml already loaded
    """
    conf = ShubConfig()
    if load_global:
        if not os.path.exists(GLOBAL_SCRAPINGHUB_YML_PATH):
            _migrate_to_global_scrapinghub_yml()
        conf.load_file(GLOBAL_SCRAPINGHUB_YML_PATH)
    if load_env and 'SHUB_APIKEY' in os.environ:
        conf.apikeys['default'] = os.environ['SHUB_APIKEY']
    if load_local:
        closest_sh_yml = closest_file('scrapinghub.yml')
        if closest_sh_yml:
            conf.load_file(closest_sh_yml)
        else:
            _migrate_and_load_scrapy_cfg(conf)
    return conf


def get_target(target, auth_required=True):
    """Load shub configuration and return target."""
    conf = load_shub_config()
    return conf.get_target(target, auth_required=auth_required)


def get_target_conf(target, auth_required=True):
    """Load shub configuration and return target."""
    conf = load_shub_config()
    return conf.get_target_conf(target, auth_required=auth_required)


def get_version():
    """Load shub configuratoin and return version."""
    conf = load_shub_config()
    return conf.get_version()


def list_targets_callback(ctx, param, value):
    """Click option callback to get targets from config, print it and exit."""
    if not value:
        return
    conf = load_shub_config()
    for name in conf.projects:
        click.echo(name)
    ctx.exit()



import unittest
from unittest import mock

from click.testing import CliRunner
from scrapinghub import ScrapinghubAPIError

from shub import schedule
from shub.exceptions import RemoteErrorException

from .utils import mock_conf


class ScheduleTest(unittest.TestCase):

    def setUp(self):
        self.runner = CliRunner()
        self.conf = mock_conf(self)

    @mock.patch('shub.schedule.schedule_spider', autospec=True)
    def test_schedules_job_if_input_is_ok(self, mock_schedule):
        proj, endpoint, apikey = self.conf.get_target('default')
        # Default
        self.runner.invoke(schedule.cli, ['spider'])
        mock_schedule.assert_called_with(
            proj, endpoint, apikey, 'spider', (), (), 2, None, (), ())
        # Other project
        self.runner.invoke(schedule.cli, ['123/spider'])
        mock_schedule.assert_called_with(
            123, endpoint, apikey, 'spider', (), (), 2, None, (), ())
        # Other endpoint
        proj, endpoint, apikey = self.conf.get_target('vagrant')
        self.runner.invoke(schedule.cli, ['vagrant/spider'])
        mock_schedule.assert_called_with(
            proj, endpoint, apikey, 'spider', (), (), 2, None, (), ())
        # Other project at other endpoint
        self.runner.invoke(schedule.cli, ['vagrant/456/spider'])
        mock_schedule.assert_called_with(
            456, endpoint, apikey, 'spider', (), (), 2, None, (), ())

    @mock.patch('shub.schedule.ScrapinghubClient', autospec=True)
    def test_schedule_invalid_spider(self, mock_client):
        mock_proj = mock_client.return_value.get_project.return_value
        mock_proj.jobs.run.side_effect = ScrapinghubAPIError('')
        with self.assertRaises(RemoteErrorException):
            schedule.schedule_spider(1, 'https://endpoint/api/',
                                     'FAKE_API_KEY', 'fake_spider')

    @mock.patch('shub.schedule.ScrapinghubClient', autospec=True)
    def test_schedule_spider_calls_project_jobs_run(self, mock_client):
        mock_proj = mock_client.return_value.get_project.return_value
        schedule.schedule_spider(1, 'https://endpoint/api/',
                                 'FAKE_API_KEY', 'fake_spider')
        self.assertTrue(mock_proj.jobs.run)

    @mock.patch('shub.schedule.ScrapinghubClient', autospec=True)
    def test_forwards_args_and_settings(self, mock_client):
        mock_proj = mock_client.return_value.get_project.return_value
        self.runner.invoke(
            schedule.cli,
            "testspider -s SETT=99 -a ARG=val1 --set SETTWITHEQUAL=10=10 "
            "--argument ARGWITHEQUAL=val2=val2".split(' '),
        )
        job_args = mock_proj.jobs.run.call_args[1]['job_args']
        self.assertLessEqual(
            {'ARG': 'val1', 'ARGWITHEQUAL': 'val2=val2'}.items(),
            job_args.items(),
        )
        job_settings = mock_proj.jobs.run.call_args[1]['job_settings']
        self.assertEqual(
            {'SETT': '99', 'SETTWITHEQUAL': '10=10'},
            job_settings,
        )

    @mock.patch('shub.schedule.ScrapinghubClient', autospec=True)
    def test_forwards_tags(self, mock_client):
        mock_proj = mock_client.return_value.get_project.return_value
        self.runner.invoke(schedule.cli, 'testspider -t tag1 -t tag2 --tag tag3'.split())
        call_kwargs = mock_proj.jobs.run.call_args[1]
        assert call_kwargs['add_tag'] == ('tag1', 'tag2', 'tag3')

    @mock.patch('shub.schedule.ScrapinghubClient', autospec=True)
    def test_forwards_priority(self, mock_client):
        mock_proj = mock_client.return_value.get_project.return_value
        # short option name
        self.runner.invoke(schedule.cli, 'testspider -p 3'.split())
        call_kwargs = mock_proj.jobs.run.call_args[1]
        assert call_kwargs['priority'] == 3
        # long option name
        self.runner.invoke(schedule.cli, 'testspider --priority 1'.split())
        call_kwargs = mock_proj.jobs.run.call_args[1]
        assert call_kwargs['priority'] == 1

    @mock.patch('shub.schedule.ScrapinghubClient', autospec=True)
    def test_forwards_units(self, mock_client):
        mock_proj = mock_client.return_value.get_project.return_value
        # no units specified
        self.runner.invoke(schedule.cli, 'testspider'.split())
        call_kwargs = mock_proj.jobs.run.call_args[1]
        assert call_kwargs['units'] is None
        # short option name
        self.runner.invoke(schedule.cli, 'testspider -u 4'.split())
        call_kwargs = mock_proj.jobs.run.call_args[1]
        assert call_kwargs['units'] == 4
        # long option name
        self.runner.invoke(schedule.cli, 'testspider --units 3'.split())
        call_kwargs = mock_proj.jobs.run.call_args[1]
        assert call_kwargs['units'] == 3

    @mock.patch('shub.schedule.ScrapinghubClient', autospec=True)
    def test_forwards_environment(self, mock_client):
        mock_proj = mock_client.return_value.get_project.return_value
        self.runner.invoke(
            schedule.cli,
            "testspider -e VAR1=VAL1 --environment VAR2=VAL2".split(' '),
        )
        call_kwargs = mock_proj.jobs.run.call_args[1]
        self.assertLessEqual(
            {'VAR1': 'VAL1', 'VAR2': 'VAL2'}.items(),
            call_kwargs['environment'].items(),
        )



import os
import re

import click

from shub import exceptions as shub_exceptions
from shub.config import load_shub_config, list_targets_callback
from shub.image import utils
from shub.image.test import test_cmd
from shub.utils import create_scrapinghub_yml_wizard


SHORT_HELP = 'Build release image.'

HELP = """
Build command uses your Dockerfile to build an image and tag it properly.

Internally, this command is a simple wrapper to `docker build` and uses
docker daemon on your system to build an image. Also it can generate
project version for you, and locate root project directory by itself.

Image should be set via scrapinghub.yml, section "images". If version is not
provided, the tool uses VCS-based stamp over project directory (the same as
shub utils itself).
"""

BUILD_STEP_REGEX = re.compile(r'Step (\d+)/(\d+) :.*')
BUILD_SUCCESS_REGEX = re.compile(r'Successfully built ([0-9a-f]+)')


@click.command(help=HELP, short_help=SHORT_HELP)
@click.argument("target", required=False, default="default")
@click.option("-l", "--list-targets", is_flag=True, is_eager=True,
              expose_value=False, callback=list_targets_callback,
              help="List available project names defined in your config")
@click.option("-d", "--debug", help="debug mode", is_flag=True,
              callback=utils.deprecate_debug_parameter)
@click.option("-v", "--verbose", is_flag=True,
              help="stream build logs to console")
@click.option("-V", "--version", help="release version")
@click.option("-S", "--skip-tests", help="skip testing image", is_flag=True)
@click.option("-n", "--no-cache", is_flag=True,
              help="Do not use cache when building the image")
@click.option("-b", "--build-arg", multiple=True,
              help="Allow to pass build arguments to docker client.")
@click.option("-f", "--file", "filename", default='Dockerfile',
              help="Name of the Dockerfile (Default is 'PATH/Dockerfile')")
def cli(target, debug, verbose, version, skip_tests, no_cache, build_arg, filename):
    build_cmd(target, version, skip_tests, no_cache, build_arg, filename=filename)


def build_cmd(target, version, skip_tests, no_cache, build_arg, filename='Dockerfile'):
    config = load_shub_config()
    create_scrapinghub_yml_wizard(config, target=target, image=True)
    client = utils.get_docker_client()
    project_dir = utils.get_project_dir()
    image = config.get_image(target)
    image_name = utils.format_image_name(image, version)
    build_args = dict(a.split('=', 1) for a in build_arg)
    if not os.path.exists(os.path.join(project_dir, filename)):
        raise shub_exceptions.NotFoundException(
            "Dockerfile is not found and it is required because project '{}' is configured "
            "to deploy Docker images. Please add a Dockerfile that will be used to build "
            "the image and retry this command. If you want to migrate an existing Scrapy project "
            "you can use `shub image init` command to create a Dockerfile.".format(target))
    if utils.is_verbose():
        build_progress_cls = _LoggedBuildProgress
    else:
        build_progress_cls = _BuildProgress
    click.echo(f"Building {image_name}.")
    events = client.build(
        path=project_dir,
        tag=image_name,
        decode=True,
        dockerfile=filename,
        nocache=no_cache,
        rm=True,
        buildargs=build_args,
    )
    build_progress = build_progress_cls(events)
    build_progress.show()
    click.echo(f"The image {image_name} build is completed.")
    # Test the image content after building it
    if not skip_tests:
        test_cmd(target, version)


class _LoggedBuildProgress(utils.BaseProgress):
    """Visualize build progress in verbose mode.

    Output all the events received from the docker daemon.
    """
    def handle_event(self, event):
        super().handle_event(event)
        if 'stream' in event:
            self.handle_stream_event(event)

    def handle_stream_event(self, event):
        utils.debug_log("{}".format(event['stream'].rstrip()))


class _BuildProgress(_LoggedBuildProgress):
    """Visualize build progress in non-verbose mode.

    Show total progress bar.
    """

    def __init__(self, events):
        super().__init__(events)
        self.bar = utils.create_progress_bar(
            total=1,
            desc='Steps',
            # don't need rate here, let's simplify the bar
            bar_format='{l_bar}{bar}| {n_fmt}/{total_fmt}',
        )
        self.is_built = False

    def show(self):
        super().show()
        if self.bar:
            self.bar.close()
        if not self.is_built:
            raise shub_exceptions.RemoteErrorException(
                "Build image operation failed")

    def handle_stream_event(self, event):
        if BUILD_SUCCESS_REGEX.search(event['stream']):
            self.is_built = True
            return
        step_row = BUILD_STEP_REGEX.match(event['stream'])
        if not step_row:
            return
        step_id, total = (int(val) for val in step_row.groups())
        self.bar.total = max(self.bar.total, total)
        # ignore onbuild sub-steps
        if step_id > self.bar.n and self.bar.total == total:
            self.bar.update()



import os
import platform
import sys
import unittest
from unittest.mock import patch, Mock

from packaging.version import parse
from pipenv import __version__ as pipenv_version
import requests
from cleo.testers.command_tester import CommandTester
from click.testing import CliRunner

from shub import deploy
from shub.exceptions import (
    NotFoundException, ShubException, BadParameterException,
    DeployRequestTooLargeException,
)
from shub.utils import create_default_setup_py, _SETUP_PY_TEMPLATE, STDOUT_ENCODING

from .utils import AssertInvokeRaisesMixin, mock_conf

try:
    from importlib.metadata import version, PackageNotFoundError
except ImportError:
    POETRY_VERSION = None
else:
    try:
        POETRY_VERSION = parse(version("poetry"))
    except PackageNotFoundError:
        try:
            POETRY_VERSION = parse(version("poetry-core"))
        except PackageNotFoundError:
            POETRY_VERSION = None

VALID_SCRAPY_CFG = """
[settings]
default = project.settings
"""


class DeployTest(AssertInvokeRaisesMixin, unittest.TestCase):

    def setUp(self):
        self.runner = CliRunner()
        self.conf = mock_conf(self, 'shub.deploy.load_shub_config')

    def _make_project(self):
        with open('scrapy.cfg', 'w') as f:
            f.write(VALID_SCRAPY_CFG)

    @patch('shub.deploy.make_deploy_request')
    def test_detect_scrapy_project(self, mock_deploy_req):
        with self.runner.isolated_filesystem():
            self.assertInvokeRaises(NotFoundException, deploy.cli)
            self._make_project()
            result = self.runner.invoke(deploy.cli)
            self.assertEqual(0, result.exit_code)

    @patch('shub.deploy.make_deploy_request')
    def _invoke_with_project(self, args, mock_deploy_req):
        with self.runner.isolated_filesystem():
            self._make_project()
            self.runner.invoke(deploy.cli, args)
        return mock_deploy_req.call_args[0]

    def test_fallback_to_default(self):
        url, data, files, auth, _, _ = self._invoke_with_project(None)
        self.assertIn(self.conf.endpoints['default'], url)
        self.assertEqual(data, {'project': 1, 'version': 'version'})
        self.assertEqual(auth, (self.conf.apikeys['default'], ''))

    def test_with_target(self):
        url, data, files, auth, _, _ = self._invoke_with_project(('prod', ))
        self.assertIn(self.conf.endpoints['default'], url)
        self.assertEqual(data, {'project': 2, 'version': 'version'})
        self.assertEqual(auth, (self.conf.apikeys['default'], ''))

    def test_with_id(self):
        url, data, files, auth, _, _ = self._invoke_with_project(('123', ))
        self.assertIn(self.conf.endpoints['default'], url)
        self.assertEqual(data, {'project': 123, 'version': 'version'})
        self.assertEqual(auth, (self.conf.apikeys['default'], ''))

    def test_with_external_id(self):
        url, data, files, auth, _, _ = self._invoke_with_project(
            ('vagrant/456', ))
        self.assertIn(self.conf.endpoints['vagrant'], url)
        self.assertEqual(data, {'project': 456, 'version': 'version'})
        self.assertEqual(auth, (self.conf.apikeys['vagrant'], ''))

    def test_deploy_list_targets(self):
        with self.runner.isolated_filesystem():
            self._make_project()
            result = self.runner.invoke(deploy.cli, ('--list-targets',))
            assert result.exit_code == 0

    @patch('shub.deploy.deploy_cmd')
    def test_custom_deploy_disabled(self, mock_deploy_cmd):
        with self.runner.isolated_filesystem():
            self._make_project()
            self.runner.invoke(deploy.cli, ('custom1',))
        self.assertTrue(mock_deploy_cmd.called)

    @patch('shub.deploy.upload_cmd')
    def test_custom_deploy_default(self, mock_upload_cmd):
        with self.runner.isolated_filesystem():
            self._make_project()
            self.runner.invoke(deploy.cli, ('custom2',))
        self.assertEqual(mock_upload_cmd.call_args[0], ('custom2', None))

    @patch('shub.deploy.upload_cmd')
    def test_custom_deploy_by_id(self, mock_upload_cmd):
        with self.runner.isolated_filesystem():
            self._make_project()
            self.runner.invoke(deploy.cli, ('5',))
        mock_upload_cmd.assert_called_once_with('5', None)

    def test_custom_deploy_bad_registry(self):
        with self.runner.isolated_filesystem():
            self._make_project()
            self.assertInvokeRaises(BadParameterException, deploy.cli, ('custom3',))

    @patch('shub.deploy.make_deploy_request')
    def test_deploy_with_custom_setup_py(self, mock_deploy_req):
        with self.runner.isolated_filesystem():
            # This scrapy.cfg contains no "settings" section, so creating a
            # default setup.py would fail (because we can't find the settings
            # module)
            open('scrapy.cfg', 'w').close()
            # However, we already have a setup.py...
            create_default_setup_py(settings='some_module')
            # ... so shub should not fail while trying to create one
            result = self.runner.invoke(deploy.cli)
            self.assertEqual(result.exit_code, 0)

    @patch('shub.utils.requests')
    @patch('shub.utils.write_and_echo_logs')
    def test_deploy_with_single_large_file(self, mock_logs, mock_requests):
        with self.runner.isolated_filesystem():
            self._make_project()
            # patch setup_py to include package data files
            setup_py = _SETUP_PY_TEMPLATE % {'settings': 'project.settings'}
            setup_py = setup_py.rsplit('\n', 2)[0]  # drop an enclosing brace
            setup_py += '    include_package_data = True)'
            with open('setup.py', 'w') as setup_file:
                setup_file.write(setup_py)
            # create a fake package and add a large random file there
            os.mkdir('files')
            open('files/__init__.py', 'wb').close()
            with open('files/file.large', 'wb') as bigfile:
                bigfile.write(os.urandom(50 * 1024 * 1024))
            # add manifest to include the non-code file
            with open('MANIFEST.in', 'w') as manifest_f:
                manifest_f.write('include files/file.large')
            fake_response = requests.Response()
            fake_response.status_code = 200
            mock_requests.post.return_value = fake_response
            self.assertInvokeRaises(DeployRequestTooLargeException,
                                    deploy.cli)


class DeployFilesTest(unittest.TestCase):
    def setUp(self):
        self.runner = CliRunner()
        self.request = patch('shub.deploy.make_deploy_request').start()
        self.addCleanup(patch.stopall)

    def _deploy(self, main_egg='./main.egg', req='./requirements.txt',
                eggs=None):
        if eggs is None:
            eggs = ['./1.egg', './2.egg']

        deploy._upload_egg(
            'endpoint',
            main_egg,
            '1', 'version', 'auth',
            False, False,
            requirements_file=req,
            eggs=eggs,
        )
        files = {}
        for name, file in self.request.call_args[0][2]:
            files.setdefault(name, []).append(
                file if isinstance(file, str)
                else file.read().decode('utf-8'))

        return files

    def test_correct_files(self):
        with self.runner.isolated_filesystem():
            with open('./main.egg', 'w') as f:
                f.write('main content')
            with open('./requirements.txt', 'w') as f:
                f.write('requirements content')
            with open('./1.egg', 'w') as f:
                f.write('1.egg content')
            with open('./2.egg', 'w') as f:
                f.write('2.egg content')
            files = self._deploy()

        self.assertEqual(files['egg'][0], 'main content')
        self.assertEqual(files['requirements'][0], 'requirements content')
        self.assertEqual(files['eggs'][0], '1.egg content')
        self.assertEqual(files['eggs'][1], '2.egg content')

    def test_no_egg(self):
        with self.runner.isolated_filesystem():
            with open('./main.egg', 'w') as f:
                f.write('main content')
            with open('./requirements.txt', 'w') as f:
                f.write('requirements content')
            with open('./1.egg', 'w') as f:
                f.write('1.egg content')

            with self.assertRaises(ShubException) as cm:
                self._deploy()

            self.assertEqual(
                cm.exception.message,
                'No such file or directory ./2.egg',
            )

    def test_no_requirements(self):
        with self.runner.isolated_filesystem():
            with open('./main.egg', 'w') as f:
                f.write('main content')
            with open('./1.egg', 'w') as f:
                f.write('1.egg content')
            with open('./2.egg', 'w') as f:
                f.write('2.egg content')

            with self.assertRaises(ShubException) as cm:
                self._deploy()

            self.assertEqual(
                cm.exception.message,
                'No such file or directory ./requirements.txt',
            )

    def test_egg_glob_pattern(self):
        with self.runner.isolated_filesystem():
            with open('./main.egg', 'w') as f:
                f.write('main content')
            with open('./a1.egg', 'w') as f:
                f.write('a1.egg content')
            with open('./a2.egg', 'w') as f:
                f.write('a2.egg content')
            with open('./b3.egg', 'w') as f:
                f.write('b3.egg content')
            files_a = self._deploy(eggs=['./a*.egg'], req=None)
            files_c = self._deploy(eggs=['./c*.egg'], req=None)
            files_all = self._deploy(eggs=['./*.egg'], req=None)
            files_main = self._deploy(eggs=['./main.egg', './*.egg'], req=None)

        self.assertEqual(len(files_a['eggs']), 2)
        self.assertIn('a1.egg content', files_a['eggs'])
        self.assertIn('a2.egg content', files_a['eggs'])
        self.assertNotIn('eggs', files_c)

        # main egg should not be added to eggs even it it matches glob pattern
        self.assertEqual(len(files_all['eggs']), 3)
        self.assertIn('a1.egg content', files_all['eggs'])
        self.assertIn('a2.egg content', files_all['eggs'])
        self.assertIn('b3.egg content', files_all['eggs'])

        # but do upload the main egg if it's directly requested
        self.assertEqual(len(files_main['eggs']), 4)
        self.assertIn('main content', files_main['eggs'])

    def test_add_sources(self):
        convert_deps_to_pip = Mock(
            side_effect=[
                './tests/requirements.txt',
                ['package==0.0.0', 'hash-package==0.0.1', 'hash-package2==0.0.1'],
            ],
        )
        _sources = (
            b'-i https://pypi.python.org/simple '
            b'--extra-index-url https://example.external-index.org/simple'
        )
        self.assertIsInstance(deploy._add_sources(convert_deps_to_pip(), _sources), str)
        self.assertIsInstance(deploy._add_sources(convert_deps_to_pip(), _sources), str)

    def pipfile_test(self, req_name):
        with self.runner.isolated_filesystem():
            with open('./main.egg', 'w') as f:
                f.write('main content')
            with open('./Pipfile.lock', 'w') as f:
                f.write("""
                {
                    "_meta": {
                        "sources": [
                            {
                                "name": "pypi",
                                "url": "https://pypi.python.org/simple",
                                "verify_ssl": true
                            },
                            {
                                "name": "external-index",
                                "url": "https://example.external-index.org/simple",
                                "verify_ssl": true
                            }
                        ]
                    },
                    "default": {
                        "package": {
                            "version": "==0.0.0"
                        },
                        "hash-package": {
                            "version": "==0.0.1",
                            "hash": "hash"
                        },
                        "hash-package2": {
                            "version": "==0.0.1",
                            "hashes": ["hash1", "hash2"]
                        },
                        "vcs-package": {
                            "git": "https://github.com/vcs/package.git",
                            "ref": "master",
                            "editable": true
                        }
                    }
                }
                """)
            with open('./1.egg', 'w') as f:
                f.write('1.egg content')
            with open('./2.egg', 'w') as f:
                f.write('2.egg content')
            files = self._deploy(req=req_name)

        reqs = set(files['requirements'][0].split('\n'))
        self.assertEqual(reqs, {
            '-i https://pypi.python.org/simple --extra-index-url https://example.external-index.org/simple',
            'package==0.0.0',
            'hash-package==0.0.1',
            'hash-package2==0.0.1',
            'git+https://github.com/vcs/package.git@master#egg=vcs-package'
            if sys.version_info < (3, 8)
            else (
                'vcs-package@ git+https://github.com/vcs/package.git@master'
                if parse(pipenv_version) < parse("2024.3.0")
                else 'vcs-package @ git+https://github.com/vcs/package.git@master'
            )
        })

    def test_pipfile_names(self):
        self.pipfile_test('Pipfile')
        self.pipfile_test('Pipfile.lock')

    def test_pipfile_lock_missing(self):
        with self.runner.isolated_filesystem():
            with open('./main.egg', 'w') as f:
                f.write('main content')
            with open('./1.egg', 'w') as f:
                f.write('1.egg content')
            with open('./2.egg', 'w') as f:
                f.write('2.egg content')

            with self.assertRaises(ShubException) as cm:
                self._deploy(req='Pipfile')

            self.assertEqual(
                cm.exception.message,
                'Please lock your Pipfile before deploying',
            )

    @patch("subprocess.check_output")
    def test_poetry_2_or_higher(self, mock_check_output):
        if not POETRY_VERSION or POETRY_VERSION < parse("2"):
            raise self.skipTest("poetry or poetry-core >=2 is not installed")

        from poetry.console.application import Application
        from poetry_plugin_export.command import ExportCommand

        with patch("poetry.packages.locker.Locker.is_fresh") as mock_is_fresh:
            with self.runner.isolated_filesystem():
                with open('./main.egg', 'w') as f:
                    f.write('main content')
                with open('./pyproject.toml', 'w') as f:
                    f.write("""
                    [project]
                    name = "test_project"
                    version = "1.2.0"
                    [tool.poetry]
                    [tool.poetry.requires-plugins]
                    poetry-plugin-export = ">=1.8"
                    """)
                with open('./poetry.lock', 'w') as f:
                    f.write("""
                    [[package]]
                    name = "package"
                    version = "0.0.0"
                    optional = false
                    python-versions = "*"
                    groups = ["main"]
                    files = []
                    [[package]]
                    name = "vcs-package"
                    version = "0.0.1"
                    optional = false
                    python-versions = "*"
                    groups = ["main"]
                    files = []
                    [package.source]
                    reference = "master"
                    type = "git"
                    url = "https://github.com/vcs/package.git"
                    [[package]]
                    name = "file-package"
                    version = "0.0.1"
                    optional = false
                    python-versions = "*"
                    groups = ["main"]
                    files = []
                    [package.source]
                    reference = ""
                    type = "file"
                    url = "/path/to/package.tar.gz"
                    [[package]]
                    name = "dir-package"
                    version = "0.0.1"
                    optional = false
                    python-versions = "*"
                    groups = ["main"]
                    files = []
                    [package.source]
                    reference = ""
                    type = "directory"
                    url = "/path/to/package"
                    [metadata.hashes]
                    package = ["hash"]
                    vcs-package = ["hash1"]
                    [metadata]
                    lock-version = "2.1"
                    """)
                with open('./1.egg', 'w') as f:
                    f.write('1.egg content')
                with open('./2.egg', 'w') as f:
                    f.write('2.egg content')

                # Skip check if lockfile is in sync with pyproject.toml
                mock_is_fresh.return_value = True

                def mock_check_output_side_effect(cmd, *args, **kwargs):
                    application = Application()
                    application.add(ExportCommand())
                    cmd = application.find("export")
                    cmd_tester = CommandTester(cmd)
                    cmd_tester.execute("--format requirements.txt")
                    self.assertIsNot(cmd_tester.status_code, 1, cmd_tester.io.fetch_error())
                    return cmd_tester.io.fetch_output().encode(STDOUT_ENCODING)

                mock_check_output.side_effect = mock_check_output_side_effect

                files = self._deploy(req="pyproject.toml")

            self.assertTrue(mock_is_fresh.called)
            self.assertTrue(mock_check_output.called)

            path_prefix = "C:/" if platform.system() == "Windows" else ""

            self.assertEqual(
                files['requirements'][0],
                (
                    f'dir-package @ file:///{path_prefix}path/to/package ; '
                    'python_version == "2.7" or python_version >= "3.4"\n'
                    f'file-package @ file:///{path_prefix}path/to/package.tar.gz ; '
                    'python_version == "2.7" or python_version >= "3.4"\n'
                    'package==0.0.0 ; '
                    'python_version == "2.7" or python_version >= "3.4"\n'
                    'vcs-package @ '
                    'git+https://github.com/vcs/package.git@master ; '
                    'python_version == "2.7" or python_version >= "3.4"'
                ),
            )

    def test_poetry_fallback(self):
        """A poetry.lock file should work even if the poetry configuration is
        invalid, with limitations like no marker support."""
        with self.runner.isolated_filesystem():
            with open('./main.egg', 'w') as f:
                f.write('main content')
            with open('./pyproject.toml', 'w') as f:
                f.write("""
                [tool.poetry]
                """)
            with open('./poetry.lock', 'w') as f:
                f.write("""
                [[package]]
                name = "package"
                version = "0.0.0"

                [[package]]
                name = "vcs-package"
                version = "0.0.1"

                [package.source]
                reference = "master"
                type = "git"
                url = "https://github.com/vcs/package.git"

                [[package]]
                name = "file-package"
                version = "0.0.1"

                [package.source]
                reference = ""
                type = "file"
                url = "/path/to/package.tar.gz"

                [[package]]
                name = "dir-package"
                version = "0.0.1"

                [package.source]
                reference = ""
                type = "directory"
                url = "/path/to/package"

                [metadata.hashes]
                package = ["hash"]
                vcs-package = ["hash1"]
                """)
            with open('./1.egg', 'w') as f:
                f.write('1.egg content')
            with open('./2.egg', 'w') as f:
                f.write('2.egg content')
            files = self._deploy(req='pyproject.toml')

        reqs = set(files['requirements'][0].split('\n'))
        self.assertEqual(reqs, {
            'package==0.0.0',
            'git+https://github.com/vcs/package.git@master#egg=vcs-package',
            '/path/to/package',
            '/path/to/package.tar.gz',
            '',
        })

    def test_poetry_lock_missing(self):
        with self.runner.isolated_filesystem():
            with open('./pyproject.toml', 'w') as f:
                f.write("""
                [tool.poetry]
                """)
            with open('./main.egg', 'w') as f:
                f.write('main content')
            with open('./1.egg', 'w') as f:
                f.write('1.egg content')
            with open('./2.egg', 'w') as f:
                f.write('2.egg content')

            with self.assertRaises(ShubException) as cm:
                self._deploy(req='pyproject.toml')

            if sys.version_info < (3, 8):
                self.assertIn(
                    "poetry executable not found",
                    cm.exception.message,
                )
            else:
                try:
                    version("poetry")
                except PackageNotFoundError:
                    self.assertIn(
                        "poetry executable not found",
                        cm.exception.message,
                    )
                else:
                    self.assertIn(
                        "The Poetry configuration is invalid",
                        cm.exception.message,
                    )



import unittest
import time
import json
from unittest import mock

from click.testing import CliRunner

from shub import items, log, requests


class JobResourceTest(unittest.TestCase):

    def setUp(self):
        self.runner = CliRunner()

    def _test_prints_objects(self, cmd_mod, resource_name):
        objects = ['Object 1', 'Object 2']
        jobid = '1/2/3'
        with mock.patch.object(cmd_mod, 'get_job', autospec=True) as mock_gj:
            # Patch job.items.iter_json() to return our objects
            mock_gj.return_value._metadata_updated = time.time()
            mock_resource = getattr(mock_gj.return_value, resource_name)
            mock_resource.iter_json.return_value = objects
            result = self.runner.invoke(cmd_mod.cli, (jobid,))
            mock_gj.assert_called_once_with(jobid)
            self.assertIn("\n".join(objects), result.output)

    def _test_forwards_follow(self, cmd_mod):
        with mock.patch.object(cmd_mod, 'get_job'), \
             mock.patch.object(cmd_mod, 'job_resource_iter', autospec=True) \
             as mock_jri:
            self.runner.invoke(cmd_mod.cli, ('1/2/3',))
            self.assertFalse(mock_jri.call_args[1]['follow'])
            self.runner.invoke(cmd_mod.cli, ('1/2/3', '-f'))
            self.assertTrue(mock_jri.call_args[1]['follow'])

    def test_items(self):
        self._test_prints_objects(items, 'items')
        self._test_forwards_follow(items)

    def test_requests(self):
        self._test_prints_objects(requests, 'requests')
        self._test_forwards_follow(requests)

    def test_log(self):
        objects = [
            {'time': 0, 'level': 20, 'message': 'message 1'},
            {'time': 1450874471000, 'level': 50, 'message': 'message 2'},
        ]
        jobid = '1/2/3'
        with mock.patch.object(log, 'get_job', autospec=True) as mock_gj:
            mock_gj.return_value._metadata_updated = time.time()
            mock_gj.return_value.logs.iter_values.return_value = objects
            result = self.runner.invoke(log.cli, (jobid,))
            mock_gj.assert_called_once_with(jobid)
            self.assertIn('1970-01-01 00:00:00 INFO message 1', result.output)
            self.assertIn('2015-12-23 12:41:11 CRITICAL message 2', result.output)
        with mock.patch.object(log, 'get_job', autospec=True) as mock_gj:
            with mock.patch.object(log, 'job_resource_iter', autospec=True) as mock_res_iter:
                mock_res_iter.return_value = [json.dumps(x) for x in objects]
                result = self.runner.invoke(log.cli, (jobid, '--json'))
                self.assertTrue(mock_res_iter.call_args[1].get('output_json'))
                for idx, line in enumerate(result.output.splitlines()):
                    self.assertEqual(json.loads(line), objects[idx])
        self._test_forwards_follow(log)

    def test_log_unicode(self):
        objects = [
            {'time': 0, 'level': 20, 'message': 'jarzębina'}
        ]
        jobid = '1/2/3'
        with mock.patch.object(log, 'get_job', autospec=True) as mock_gj:
            mock_gj.return_value._metadata_updated = time.time()
            mock_gj.return_value.logs.iter_values.return_value = objects
            result = self.runner.invoke(log.cli, (jobid,))
            mock_gj.assert_called_once_with(jobid)
            self.assertIn('1970-01-01 00:00:00 INFO jarzębina', result.output)
